from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.builder.ai import create_ai_narration


class ElementBuilder:
    @classmethod
    def get_subclasses(cls):
        """
        Returns all the existing subclasses of ElementBuilder.
        """
        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder

        return [
            ImageElementBuilder,
            AIImageElementBuilder,
            VideoElementBuilder,
            AIVideoElementBuilder,
            CustomStockElementBuilder,
            StockElementBuilder,
            SoundElementBuilder,
            MemeElementBuilder,
            YoutubeVideoElementBuilder,
            TextElementBuilder,
            PremadeElementBuilder
        ]
    
    @classmethod
    def build_narration(cls, text: str, output_filename: str):
        """
        Generates a narration file that narrates the 'text' provided and
        is stored locally as 'output_filename'. If 'text' or 
        'output_filename' fields are not provided it will raise an 
        Exception.
        """
        ElementParameterValidator.validate_string_mandatory_parameter(text)
        ElementParameterValidator.validate_string_mandatory_parameter(output_filename)

        return create_ai_narration(text, output_filename = output_filename)
    
    @classmethod
    def handle_narration_from_segment(cls, segment: dict):
        pass


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.rules.element_rules import ElementRules
from yta_general_utils.programming.error_message import ErrorMessage as BaseErrorMessage


class ErrorMessage(BaseErrorMessage):
    @classmethod
    def parameter_is_not_rules(cls, parameter_name: str):
        return f'The provided "{parameter_name}" parameter is not a valid ElementRules subclass. The valid ones are: {', '.join(ElementRules.get_subclasses().__str__)}.'
    
    @classmethod
    def parameter_is_not_builder(cls, parameter_name: str):
        return f'The provided "{parameter_name}" parameter is not a valid ElementBuilder subclass. The valid ones are: {', '.join(ElementBuilder.get_subclasses().__str__)}.'


from youtube_autonomous.segments.enums import EnhancementElementMode
from youtube_autonomous.segments.enums import SegmentType
from yta_general_utils.programming.parameter_validator import ParameterValidator
from typing import Union


class ElementRules:
    """
    Class to specify the rules that a Project Element must follow.
    The Project Element is equivalent to a Segment.
    """
    can_have_narration: bool = False
    """
    Can be narrated by setting a 'audio_narration_filename' or
    'narration_text' and 'voice'.
    """
    need_narration: bool = False
    """
    Narration is mandatory so 'audio_narration_filename' or 'voice'
    and 'narration_text' are needed.
    """
    can_have_specific_duration: bool = False
    """
    Duration of this element can be set specifically so the element
    will long as much as 'duration' field says.
    """
    need_specific_duration: bool = False
    """
    Specific 'duration' field is mandatory if it can't have narration
    or if it can but the fields needed for the narration do not exist.
    """
    can_have_text: bool = False
    """
    A 'text' field can be used to build the element in an specific
    way (determined by its type).
    """
    need_text: bool = False
    """
    The 'text' is mandatory for this element type so it must exist.
    """
    can_have_filename: bool = False
    """
    The 'filename' field can be present to load the file from that
    local stored filename.
    """
    can_have_url: bool = False
    """
    The 'url' field can be set to obtain the element from that 
    source.
    """
    need_filename_or_url: bool = False
    """
    The 'url' field or the 'filename' field are needed to be able
    to obtain the object. Priority is first 'url' then 'filename'
    if both are provided.
    """
    can_have_keywords: bool = False
    """
    The field 'keywords' can be set to look for the source in an 
    specific way or to build correctly the element (the way those
    'keywords' are used depends on the element type).
    """
    need_keywords: bool = False
    """
    The field 'keywords' is mandatory and must be set to be able
    to build this element.
    """
    can_have_more_parameters: bool = False
    """
    This field acts as a flag to let us know if the element can
    have more dynamic parameters according to the element 
    requirements. Texts or premades could have more parameters
    to be able to build them properly, and those parameters will
    be added dynamically and obtained also dynamically (they will
    be different of the static ones we have defined as the main
    fields).
    """
    can_be_segment: bool = False
    """
    This field indicates if the element with this rule can be used
    as a segment or not.
    """
    can_be_enhancement_element: bool = False
    """
    This field indicates if the element with this rule can be used
    as an enhancement element (that is added on a segment element).
    """
    valid_enhancement_modes: list[EnhancementElementMode] = []
    """
    This field indicates the valid modes for this element when
    being used as an enhancement element.
    """
    default_enhancement_mode: EnhancementElementMode = None
    """
    This field indicates the mode by default for this element when
    used as an enhancement element (if possible).
    """

    def __init__(self, can_have_narration: bool, need_narration: bool, can_have_specific_duration: bool, need_specific_duration: bool, can_have_text: bool, need_text: bool, can_have_filename: bool, can_have_url: bool, need_filename_or_url: bool, can_have_keywords: bool, need_keywords: bool, can_have_more_parameters: bool, can_be_segment: bool, can_be_enhancement_element: bool, valid_enhancement_modes: list[EnhancementElementMode], default_enhancement_mode: EnhancementElementMode):
        ParameterValidator.validate_bool_parameter('can_have_narration', can_have_narration)
        ParameterValidator.validate_bool_parameter('need_narration', need_narration)
        ParameterValidator.validate_bool_parameter('can_have_specific_duration', can_have_specific_duration)
        ParameterValidator.validate_bool_parameter('need_specific_duration', need_specific_duration)
        ParameterValidator.validate_bool_parameter('can_have_text', can_have_text)
        ParameterValidator.validate_bool_parameter('need_text', need_text)
        ParameterValidator.validate_bool_parameter('can_have_filename', can_have_filename)
        ParameterValidator.validate_bool_parameter('can_have_url', can_have_url)
        ParameterValidator.validate_bool_parameter('need_filename_or_url', need_filename_or_url)
        ParameterValidator.validate_bool_parameter('can_have_keywords', can_have_keywords)
        ParameterValidator.validate_bool_parameter('need_keywords', need_keywords)
        ParameterValidator.validate_bool_parameter('can_have_more_parameters', can_have_more_parameters)
        ParameterValidator.validate_bool_parameter('can_be_segment', can_be_segment)
        ParameterValidator.validate_bool_parameter('can_be_enhancement_element', can_be_enhancement_element)
        # TODO: Validate 'valid_enhancement_modes' and 'default_enhancement_mode'
        
        self.can_have_narrration = can_have_narration
        self.need_narration = need_narration
        self.can_have_specific_duration = can_have_specific_duration
        self.need_specific_duration = need_specific_duration
        self.can_have_text = can_have_text
        self.need_text = need_text
        self.can_have_filename = can_have_filename
        self.can_have_url = can_have_url
        self.need_filename_or_url = need_filename_or_url
        self.can_have_keywords = can_have_keywords
        self.need_keywords = need_keywords
        self.can_have_more_parameters = can_have_more_parameters

        self.can_be_segment = can_be_segment
        self.can_be_enhancement_element = can_be_enhancement_element
        self.valid_enhancement_modes = valid_enhancement_modes
        self.default_enhancement_mode = default_enhancement_mode

    @classmethod
    def get_subclasses(cls):
        """
        Returns all the existing subclasses of the ElementRules.
        """
        from youtube_autonomous.elements.rules.ai_image_element_rules import AIImageElementRules
        from youtube_autonomous.elements.rules.ai_video_element_rules import AIVideoElementRules
        from youtube_autonomous.elements.rules.image_element_rules import ImageElementRules
        from youtube_autonomous.elements.rules.video_element_rules import VideoElementRules
        from youtube_autonomous.elements.rules.stock_element_rules import StockElementRules
        from youtube_autonomous.elements.rules.custom_stock_element_rules import CustomStockElementRules
        from youtube_autonomous.elements.rules.meme_element_rules import MemeElementRules
        from youtube_autonomous.elements.rules.sound_element_rules import SoundElementRules
        from youtube_autonomous.elements.rules.youtube_video_element_rules import YoutubeVideoElementRules
        from youtube_autonomous.elements.rules.text_element_rules import TextElementRules
        from youtube_autonomous.elements.rules.premade_element_rules import PremadeElementRules
        from youtube_autonomous.elements.rules.effect_element_rules import EffectElementRules
        from youtube_autonomous.elements.rules.greenscreen_element_rules import GreenscreenElementRules

        return [
            AIImageElementRules,
            AIVideoElementRules,
            ImageElementRules,
            VideoElementRules,
            StockElementRules,
            CustomStockElementRules,
            MemeElementRules,
            SoundElementRules,
            YoutubeVideoElementRules,
            TextElementRules,
            PremadeElementRules,
            EffectElementRules,
            GreenscreenElementRules
        ]
    
    @classmethod
    def get_rules_class_by_type(cls, type: Union[SegmentType, str]):
        if not type:
            raise Exception('No "type" provided.')
        
        if not isinstance(type, (SegmentType, str)):
            raise Exception('The "type" parameter provided is not a SegmentType nor a string.')
        
        if isinstance(type, str):
            if not SegmentType.is_valid(type):
                raise Exception('The "type" parameter provided is not a valid SegmentType string value.')
            
            type = SegmentType(type)

        from youtube_autonomous.elements.rules.ai_image_element_rules import AIImageElementRules
        from youtube_autonomous.elements.rules.ai_video_element_rules import AIVideoElementRules
        from youtube_autonomous.elements.rules.image_element_rules import ImageElementRules
        from youtube_autonomous.elements.rules.video_element_rules import VideoElementRules
        from youtube_autonomous.elements.rules.stock_element_rules import StockElementRules
        from youtube_autonomous.elements.rules.custom_stock_element_rules import CustomStockElementRules
        from youtube_autonomous.elements.rules.meme_element_rules import MemeElementRules
        from youtube_autonomous.elements.rules.sound_element_rules import SoundElementRules
        from youtube_autonomous.elements.rules.youtube_video_element_rules import YoutubeVideoElementRules
        from youtube_autonomous.elements.rules.text_element_rules import TextElementRules
        from youtube_autonomous.elements.rules.premade_element_rules import PremadeElementRules
        from youtube_autonomous.elements.rules.effect_element_rules import EffectElementRules
        from youtube_autonomous.elements.rules.greenscreen_element_rules import GreenscreenElementRules

        if type == SegmentType.MEME:
            return MemeElementRules
        elif type == SegmentType.AI_IMAGE:
            return AIImageElementRules
        elif type == SegmentType.AI_VIDEO:
            return AIVideoElementRules
        elif type == SegmentType.IMAGE:
            return ImageElementRules
        elif type == SegmentType.VIDEO:
            return VideoElementRules
        elif type == SegmentType.STOCK:
            return StockElementRules
        elif type == SegmentType.CUSTOM_STOCK:
            return CustomStockElementRules
        elif type == SegmentType.SOUND:
            return SoundElementRules
        elif type == SegmentType.YOUTUBE_VIDEO:
            return YoutubeVideoElementRules
        elif type == SegmentType.TEXT:
            return TextElementRules
        elif type == SegmentType.PREMADE:
            return PremadeElementRules
        elif type == SegmentType.EFFECT:
            return EffectElementRules
        elif type == SegmentType.GREENSCREEN:
            return GreenscreenElementRules


from youtube_autonomous.elements.rules.ai_image_element_rules import AIImageElementRules
from youtube_autonomous.elements.rules.image_element_rules import ImageElementRules
from youtube_autonomous.elements.rules.meme_element_rules import MemeElementRules
from youtube_autonomous.elements.rules.ai_video_element_rules import AIVideoElementRules
from youtube_autonomous.elements.rules.video_element_rules import VideoElementRules
from youtube_autonomous.elements.rules.sound_element_rules import SoundElementRules
from youtube_autonomous.elements.rules.custom_stock_element_rules import CustomStockElementRules
from youtube_autonomous.elements.rules.stock_element_rules import StockElementRules
from youtube_autonomous.elements.rules.youtube_video_element_rules import YoutubeVideoElementRules
from youtube_autonomous.elements.rules.text_element_rules import TextElementRules
from youtube_autonomous.elements.rules.premade_element_rules import PremadeElementRules
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from typing import Union


class ProjectElement:
    @property
    def builder(self):
        return self._builder
    
    @builder.setter
    def builder(self, builder: Union[ImageElementBuilder, AIImageElementBuilder, VideoElementBuilder, AIVideoElementBuilder, CustomStockElementBuilder, StockElementBuilder, SoundElementBuilder, MemeElementBuilder, YoutubeVideoElementBuilder, TextElementBuilder, PremadeElementBuilder]):
        ElementParameterValidator.validate_builder(builder)
        
        self._builder = builder

    @property
    def rules(self):
        return self._rules
    
    @rules.setter
    def rules(self, rules: Union[AIImageElementRules, ImageElementRules, MemeElementRules, AIVideoElementRules, VideoElementRules, SoundElementRules, CustomStockElementRules, StockElementRules, YoutubeVideoElementRules, TextElementRules, PremadeElementRules]):
        ElementParameterValidator.validate_rules(rules)

        self._rules = rules

    @property
    def rules_checker(self):
        return self._rules_checker
    
    @rules_checker.setter
    def rules_checker(self, rules_checker: RulesChecker):
        # TODO: move this to 'ElementParameterValidator'
        if not isinstance(rules_checker, RulesChecker):
            raise Exception(f'The provided "rules_checker" parameter is not one of the following instances: {", ".join([RulesChecker.__str__])}')
        
        self._rules_checker = rules_checker

    def __init__(self, rules: Union[AIImageElementRules, ImageElementRules, MemeElementRules, AIVideoElementRules, VideoElementRules, SoundElementRules, CustomStockElementRules, StockElementRules, YoutubeVideoElementRules, TextElementRules, PremadeElementRules], builder: Union[ImageElementBuilder, AIImageElementBuilder, VideoElementBuilder, AIVideoElementBuilder, CustomStockElementBuilder, StockElementBuilder, SoundElementBuilder, MemeElementBuilder, YoutubeVideoElementBuilder, TextElementBuilder, PremadeElementBuilder]):
        self.rules = rules
        self.builder = builder
        self.rules_checker = RulesChecker(rules)

        # TODO: Add 'segment' but less strict (?)
    
class MemeElement(ProjectElement):
    def __init__(self):
        super().__init__(rules = MemeElementRules, builder = MemeElementBuilder)

    def build(self):
        self.builder.build()



from youtube_autonomous.segments.builder.youtube.youtube_downloader import YoutubeDownloader
from youtube_autonomous.segments.builder.stock.stock_downloader import StockDownloader
from youtube_autonomous.segments.enums import SegmentType
from youtube_autonomous.segments.builder.config import MAX_DURATION_PER_IMAGE, MIN_DURATION_PER_IMAGE, MAX_DURATION_PER_YOUTUBE_SCENE
from youtube_autonomous.segments.builder.ai import create_ai_image
from yta_multimedia.image.edition.resize import resize_image_scaling
from yta_multimedia.video.dimensions import rescale_video
from yta_multimedia.video.generation import generate_video_from_image
from yta_multimedia.video.generation.google.google_search import GoogleSearch
from yta_multimedia.video.generation.google.youtube_search import YoutubeSearch
from yta_multimedia.video.generation.manim.classes.text.text_word_by_word_manim_animation import TextWordByWordManimAnimation
from yta_general_utils.downloader.image import download_image
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.logger import print_in_progress
from moviepy.editor import VideoFileClip, concatenate_videoclips


class SegmentContentBuilder:
    """
    Class that is able to build the content of a project segment with
    the parameters that are set in the segment.
    """
    def __init__(self):
        # TODO: These objects below are special ones of this project
        # to handle Youtube and Stock elements. I cannot import
        # directly from libraries to use them because it has more
        # logic
        self.youtube_handler = YoutubeDownloader(True)
        self.stock_handler = StockDownloader(True)

    # TODO: Continue
    def build_segment(self, segment):
        """
        Builds a moviepy VideoFileClip based on the provided 'segment'
        properties.
        """
        video = None

        if segment.type == SegmentType.IMAGE.value:
            video = self.__build_image(segment)
        elif segment.type == SegmentType.AI_IMAGE.value:
            video = self.__build_ai_image(segment)
        elif segment.type == SegmentType.CUSTOM_STOCK.value:
            video = self.__build_custom_stock(segment)
        elif segment.type == SegmentType.STOCK.value:
            video = self.__build_stock(segment)
        elif segment.type == SegmentType.YOUTUBE_VIDEO.value:
            video = self.__build_youtube_video(segment)
        elif segment.type == SegmentType.TEXT.value:
            video = self.__build_text_video(segment)
        elif segment.type == SegmentType.MEME.value:
            video = self.__build_meme_video(segment)
        # Accepted premades below
        # TODO: Implement the 'premade' type and the 'subtype' field
        # to manage these kind of premades
        elif segment.type in SegmentType.get_premade_types().values():
            video = self.__build_premade_video(segment)
        else:
            # TODO: This should not happend because of the validator but...
            raise Exception(f'Unexpected segment type "{str(segment.type)}" for content creation.')

        return video

    def __build_image(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.IMAGE) video and
        returns it as a clip.

        This method will download the image in the 'segment.url' parameter, resize
        it to fit a 1920x1080 screen size, and build the video with a duration of
        'segment.duration' seconds.

        TODO: Review this explanation
        """
        # TODO: We maybe do not need this because of alternative creations which are
        # done when the main type is not available because of any external reason
        # if segment.type != SegmentType.IMAGE:
        #     raise Exception('The given segment is not a SegmentType.IMAGE type.')

        # TODO: Try to do all this process in memory, writting not the image
        image_filename = download_image(segment.url)

        # Resize image to fit the screen
        tmp_image_filename = create_temp_filename(f'image{get_file_extension(image_filename)}')
        resize_image_scaling(image_filename, 1920, 1080, tmp_image_filename)

        # TODO: Apply and Effect (this need work)
        video = generate_video_from_image(tmp_image_filename, segment.calculated_duration, None)

        return video
    
    def __build_ai_image(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.AI_IMAGE) 
        video and returns it as a clip.

        This method will use the segment keywords to generate one (or more)
        ai images for the moviepy video clip that is returned.
        """
        # We need to know the amount of ai images we need to get the final video
        # because we cannot use only one image for a long time
        # TODO: Implement a 'strategy' field to know if only 1 image or more than
        # one. By now 'more than one' is default
        images = []
        if segment.calculated_duration > MAX_DURATION_PER_IMAGE:
            number_of_images = int(segment.calculated_duration / MAX_DURATION_PER_IMAGE)
            for i in range(number_of_images):
                images.append({
                    'keywords': segment.keywords,
                    'duration': MAX_DURATION_PER_IMAGE
                })

            remaining_duration = segment.calculated_duration % MAX_DURATION_PER_IMAGE
            if remaining_duration <= MIN_DURATION_PER_IMAGE:
                # We won't create an image animation that is too short, we
                # will make the previous one longer
                images[len(images) - 1]['duration'] += remaining_duration
            else:
                images.append({
                    'keywords': segment.keywords,
                    'duration': remaining_duration
                })
        else:
            images.append({
                'keywords': segment.keywords,
                'duration': segment.calculated_duration
            })

        # Let's build the final video
        videos = []
        for image in images:
            image_filename = create_ai_image(image.keywords)
            # Resize image to fit the screen
            tmp_image_filename = create_temp_filename(f'image{get_file_extension(image_filename)}')
            resize_image_scaling(image_filename, 1920, 1080, tmp_image_filename)

            # TODO: Apply and Effect (this need work)
            videos.append(generate_video_from_image(tmp_image_filename, image.duration, None))

        video = concatenate_videoclips(videos)

        return video

    def __build_custom_stock(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.CUSTOM_STOCK) 
        video and returns it as a clip.

        This method will use the segment keywords to look for custom stock
        videos for the moviepy video clip that is returned.
        """
        videos = []
        do_use_youtube = True # to stop searching in Youtube if no videos available
        accumulated_duration = 0
        while accumulated_duration < segment.calculated_duration:
            downloaded_filename = None
            if do_use_youtube:
                # We try to download if from Youtube
                print_in_progress('Downloading youtube stock video')
                youtube_stock_video = self.youtube_handler.get_stock_video(segment.keywords)
                if youtube_stock_video:
                    downloaded_filename = youtube_stock_video.download_with_audio(output_filename = create_temp_filename('youtube.mp4'))
                    if downloaded_filename:
                        self.youtube_handler.add_ignored_id(youtube_stock_video.id)

            if not downloaded_filename:
                # Not found or not searching on Youtube, so build 'stock'
                print_in_progress('Downloading stock video (youtube not found)')
                do_use_youtube = False
                video = self.build_stock_video_content_clip_from_segment(segment)
            else:
                video = VideoFileClip(downloaded_filename)

            accumulated_duration += video.duration

            if accumulated_duration > segment.calculated_duration:
                end = video.duration - (accumulated_duration - segment.calculated_duration)
                start = 0
                if youtube_stock_video:
                    if youtube_stock_video.key_moment != 0:
                        # Ok, lets use that key moment as the center of our video
                        start = youtube_stock_video.key_moment - (end / 2)
                        end = youtube_stock_video.key_moment + (end / 2)
                        if start < 0:
                            end += abs(0 - start)
                            start = 0
                        if end > video.duration:
                            start -= abs(end - video.duration)
                            end = video.duration
                    video = video.subclip(start, end)
                else:
                    video = video.subclip(start, end)

            videos.append(video)

        return concatenate_videoclips(videos)
    
    def __build_stock(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.STOCK) 
        video and returns it as a clip.

        This method will use the segment keywords to look for stock videos 
        for the moviepy video clip that is returned.
        """
        videos = []
        accumulated_duration = 0
        while accumulated_duration < segment.calculated_duration:
            print_in_progress('Downloading stock video')
            # TODO: Make this force 1920x1080 resolution
            downloaded_filename = self.stock_handler.download_video(segment.keywords, True)

            if not downloaded_filename:
                # No stock videos available, lets build with AI
                video = self.__build_ai_image(segment)
            else:
                video = rescale_video(downloaded_filename)
                # TODO: Maybe 'resize_video' instead of 'rescale_video' (?)

            accumulated_duration += video.duration
            # Last clip must be cropped to fit the expected duration
            if accumulated_duration > segment.calculated_duration:
                video = video.subclip(0, video.duration - (accumulated_duration - segment.calculated_duration))
            # TODO: I'm forcing 1920, 1080 here but it must come from Pexels
            videos.append(video)

        return concatenate_videoclips(videos)

    # TODO: This method should be refactored as it is not worth I think
    def __build_youtube_video(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.YOUTUBE_VIDEO) 
        video and returns it as a clip.

        This method will use the segment url to look for that Youtube video and
        make a custom moviepy video clip that is returned.
        """
        # TODO: This functionality must go to 'yta_multimedia' or, at
        # least, to 'youtube_handler' because I will have different
        # strategies to build segment videos from Youtube videos so I
        # need this method clean to choose a method according to the 
        # strategy without too much code here :)

        # Get the youtube video
        youtube_video = self.youtube_handler.get_video(segment.url)

        if not youtube_video.is_available():
            raise Exception('The youtube video with url "' + str(segment.url) + '" is not available.')

        scenes_number = segment.calculated_duration / MAX_DURATION_PER_YOUTUBE_SCENE
        if (segment.calculated_duration + MAX_DURATION_PER_YOUTUBE_SCENE) > 0:
            scenes_number += 1
        scene_duration = segment.calculated_duration / scenes_number

        youtube_video_scenes = []
        if youtube_video.heatmap:
            youtube_video_scenes = youtube_video.get_hottest_scenes(scenes_number, scene_duration)
        else:
            youtube_video_scenes = youtube_video.get_scenes(scenes_number, scene_duration)

        # Now we have all scenes, subclip the youtube clip
        youtube_clip = VideoFileClip(self.youtube_handler.download_this_video(youtube_video))
        scene_clips = []
        for scene in youtube_video_scenes:
            scene_clips.append(youtube_clip.subclip(scene['start'], scene['end']))

        return concatenate_videoclips(scene_clips)
    
    def __build_text_video(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.TEXT) 
        video and returns it as a clip.

        This method will use the segment text create a custom text animation 
        moviepy video clip that is returned.
        """
        # TODO: I need to add new fields like 'strategy', 'subtype', 'background'
        # or things that make me being able to customize the text video creation
        # and later, if they are not provided, just randomize them in code but
        # from a wide group of options
        
        # TODO: We need to check if this text must be build from the transcription
        # or if it is a plain text with a whole and specific duration, with or 
        # without audio... Little more complex than previous ones
        temp_filename = create_temp_filename('manim.mp4')
        # OPTIONS BELOW
        if segment.audio_narration_filename:
            # Text without duration but it has audio narration file
            duration = segment.audio_clip.duration
            transcription = segment.transcription
            # TODO: We need a strategy to build this

            # TODO: Please, document what does whisper returns inside
            # the object and in 'transcription' property in 'segment.py'
            # file to not forget it
            print(transcription)
            
            # TODO: This is just to test, use one synchronized with 
            # transcription
            text = ' '.join(d['word'] for d in transcription)
            temp_filename = TextWordByWordManimAnimation().generate(text, duration, temp_filename)
        # Text without duration but its narrated by system
        elif segment.text and segment.voice:
            duration = segment.audio_clip.duration
            transcription = segment.transcription
            # TODO: We need a strategy to build this

            # TODO: This is just to test, use one synchronized with 
            # transcription
            text = ' '.join(d['word'] for d in transcription)
            temp_filename = TextWordByWordManimAnimation().generate(text, duration, temp_filename)
        else:
            if not segment.text:
                raise Exception('No "text" in segment to be able to build the text video.')
            
            # TODO: I think any other option is avoided by the validator, but
            # confirm it and remove this comment then
            
            # Text with duration (and music maybe) but no narration
            duration = segment.duration
            text = segment.text
            # TODO: We need a strategy to build this

            temp_filename = TextWordByWordManimAnimation().generate(text, duration, temp_filename)
        
        return VideoFileClip(temp_filename)
    
    def __build_meme_video(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType.MEME) video 
        and returns it as a clip.

        This method will use the segment keywords for a custom meme to create 
        the moviepy video clip that is returned.
        """
        self.youtube_handler.deactivate_ignore_repeated()
        temp_filename = self.youtube_handler.download_meme_video(segment.keywords, True, True)
        self.youtube_handler.activate_ignore_repeated()

        # TODO: Look for a better strategy (?)
        if not temp_filename:
            raise Exception('No meme found with the given "keywords": ' + str(segment.keywords) + '.')
        

        return VideoFileClip(temp_filename)
    
    def __build_premade_video(self, segment):
        """
        Builds the provided 'segment' (that must be a SegmentType of a premade) 
        video and returns it as a clip.

        This method will use the segment parameters to build a custom premade 
        moviepy video clip that is returned.
        """
        video = None

        if segment.type == SegmentType.GOOGLE_SEARCH.value:
            video = GoogleSearch(self.keywords).generate()
        elif segment.type == SegmentType.YOUTUBE_SEARCH.value:
            video = YoutubeSearch(self.keywords).generate()

        return video



from yta_stock_downloader.videos.pexels import download_first as download_pexels_first_video, download_random as download_pexels_random_video
from yta_stock_downloader.videos.pixabay import download_first as download_pixabay_first_video
from yta_general_utils.temp import create_temp_filename


class StockDownloader:
    """
    Singleton class.

    This object simplifies the access to our Stock provider platforms.
    It uses our stock library to connecto with different stock 
    platforms directly to download the content.

    This object is used to download stock videos to build the 
    main images that are shown in the videos.
    """
    __instance = None

    def __new__(cls, ignore_repeated = True):
        if not StockDownloader.__instance:
            StockDownloader.__instance = object.__new__(cls)
        
        return StockDownloader.__instance

    def __init__(self, ignore_repeated = True):
        if not hasattr(self, 'ignore_repeated'):
            self.ignore_repeated = ignore_repeated
            self.pexels_used_video_ids = []
            self.pexels_used_image_ids = []
            self.pexels_ignore_ids = []
            self.pixabay_used_video_ids = []
            self.pixabay_used_image_ids = []
            self.pixabay_ignore_ids = []

    def download_video(self, keywords: str, do_randomize: bool = False, output_filename: str = create_temp_filename('stock.mp4')):
        """
        Searches the provided 'keywords' in the different stock platforms we have 
        looking for videos. This method will download, if any video is found, a 
        video that could be the first one or a random one if 'do_randomize' is
        True.

        This method will return the locally stored filename when downloaded, or
        None if no video is found.
        """
        if not keywords:
            raise Exception('No "keywords" provided.')
        
        if not output_filename:
            output_filename = create_temp_filename('stock.mp4')

        # TODO: What about the sound?
        if do_randomize:
            download = download_pexels_random_video(keywords, self.pexels_used_video_ids, output_filename)
            if not download:
                # TODO: Implement 'download_pixabay_random_video' (check task)
                download = download_pixabay_first_video(keywords, output_filename)
        else:
            download = download_pexels_first_video(keywords, self.pexels_used_video_ids, output_filename)
            if not download:
                download = download_pixabay_first_video(keywords, output_filename)

        if not download:
            return None

        if self.ignore_repeated:
            if isinstance(download, str):
                self.pixabay_used_video_ids.append(download)
            else:
                self.pexels_used_video_ids.append(download['id'])

        if isinstance(download, str):
            output_filename = download
        else:
            output_filename = download['output_filename']

        return output_filename
    


from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.segments.validation.segment_validator import SegmentValidator
from youtube_autonomous.shortcodes.objects.shortcode_tag import ShortcodeTag, ShortcodeType
from youtube_autonomous.project import Project
from youtube_autonomous.segments.builder.config import DEFAULT_SEGMENT_PARTS_FOLDER, DEFAULT_PROJECTS_OUTPUT_FOLDER
from youtube_autonomous.shortcodes.shortcode_parser import ShortcodeParser
from yta_general_utils.temp import clean_temp_folder
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.path import create_file_abspath
from yta_general_utils.file.reader import read_json_from_file
from yta_general_utils.logger import print_completed, print_in_progress
from typing import Union


class YoutubeAutonomous:
    __database_handler: DatabaseHandler = None
    """
    Object to interact with the database and get and create projects.

    _This parameter is not manually set by the user._
    """
    __segment_validator: SegmentValidator = None
    """
    Object to validate the segments we want to use to build a project.

    _This parameter is not manually set by the user._
    """
    __segments_abspath: Union[str, None] = None
    """
    The absolute path that points to the segments folder in which this
    software will store the segments related content parts that are
    created during the whole segment building process. This folder 
    allows us recovering a non-finished project in which we generated
    some segments in the past but not all of them.

    _This parameter is not manually set by the user._
    """
    __projects_output_abspath: Union[str, None] = None
    """
    The absolute path that points to the projects output folder in which
    this software will store the final videos generated.

    _This parameter is not manually set by the user._
    """

    def __init__(self):
        """
        Initializes the object and creates the segment part building
        folder in which segment parts will be stored.
        """
        self.__database_handler = DatabaseHandler()
        self.__segment_validator = SegmentValidator()

        self.__segments_abspath = get_project_abspath() + DEFAULT_SEGMENT_PARTS_FOLDER + '/'
        self.__projects_output_abspath = get_project_abspath() + DEFAULT_PROJECTS_OUTPUT_FOLDER + '/'

        # We force to create the folder if it doesn't exist
        create_file_abspath(self.__segments_abspath + 'toforce')
        create_file_abspath(self.__projects_output_abspath + 'toforce')

    def purge(self, do_remove_segments_files = False):
        """
        Cleans the temporary folder removing all previous generated 
        temporary files, and also the segment files if the
        'do_remove_segments_files' parameter is set to True.
        """
        clean_temp_folder()
        if do_remove_segments_files:
            # TODO: Remove all files in self.__segments_abspath folder
            pass

    def check_config(self):
        # TODO: Check that he config is ok
        pass
     
    def insert_project_in_database_from_file(self, filename: str):
        """
        Reads the provided project content 'filename' and creates a new 
        project in the database if the provided 'filename' contains a new
        project and is valid. If the information belongs to an already
        registered project, it will raise an exception.

        This method returns the new stored project mongo ObjectId if 
        successfully stored, or raises an Exception if anything went wrong.
        """
        if not filename:
            raise Exception('We need a project content "filename" to create a new project.')
        
        if not isinstance(filename, str):
            raise Exception('The "filename" parameter provided is a str.')
        
        json_data = self.__read_project_from_file(filename)

        # If a project with the same content exists, it is the same project
        db_project = self.__database_handler.get_database_project_from_json(json_data)
        if db_project:
            raise Exception('There is an existing project in the database with the same content.')

        # TODO: Process the information to store it parsed
        json_data = self.check_and_prepare_data(json_data)

        db_project = self.__database_handler.insert_project(json_data)

        print_completed('Project created in database with ObjectId = "' + str(db_project['_id']) + '"')

        return str(db_project['_id'])
    
    def validate_data(self, json_data: dict):
        segments = json_data.get('segments', [])

        if not segments:
            raise Exception('No "segments" in the provided "json_data" parameter.')

        # TODO: Check main structure and validate fields
        # TODO: This shortcode parser must have the definitive shortcodes
        # TODO: Shortcode handlers must be working, check Notion (https://www.notion.so/ShortcodeParser-personalizado-10ff5a32d462804d9253e9cb782e5540?pvs=4)
        shortcode_parser = ShortcodeParser([ShortcodeTag('meme', ShortcodeType.SIMPLE)])

        from youtube_autonomous.elements.rules.element_rules import ElementRules
        from youtube_autonomous.segments.enums import SegmentType

        for segment in segments:
            type = json_data.get('type', '')

            if not type or not SegmentType.is_valid(type):
                raise Exception('The "type" provided is not valid.')

            # 1. If 'text' have shortcodes
            # We use an empty shortcode parser to accept no shortcodes
            # so any shortcode found will raise an Exception
            ShortcodeParser([]).parse(segment.get('text', ''))

            # 2. If 'narration_text' have invalid shortcodes
            # The final shortcode will parse only the accepted shortcodes
            # so if any unknown shortcode appears it will raise an Exception
            shortcode_parser.parse(segment.get('narration_text', ''))

            # 3. If 'data.effects' are invalid
            from youtube_autonomous.elements.rules.effect_element_rules import EffectElementRules
            from youtube_autonomous.elements.rules.greenscreen_element_rules import GreenscreenElementRules
            from youtube_autonomous.elements.rules.rules_checker import RulesChecker
            # TODO: This must be const or something like that
            accepted_fields = ['greenscreens', 'effects']
            # TODO: Delete any other field that is not one of above list
            for enhancement in segment.get('enhancements', []):
                effects = enhancement.get('effects', [])
                for effect in effects:
                    RulesChecker(EffectElementRules()).check_this_need_rules(effect)

                greenscreens = enhancement.get('greenscreens', [])
                for greenscreen in greenscreens:
                    RulesChecker(GreenscreenElementRules()).check_this_need_rules(greenscreen)

                # TODO: We should check any other thing that is here (frames, etc.)

            # 4. If doesn't have necessary fields for its type
            rules_checker = RulesChecker(ElementRules.get_rules_class_by_type(type))
            rules_checker.check_this_need_rules(segment)

    
    def check_and_prepare_data(self, json_data: dict):
        self.validate_data(json_data)

        shortcode_parser = ShortcodeParser()
        for segment in json_data['segments']:
            # 1. Parse shortcodes of 'narration_text' if existing
            if segment['narration_text']:
                shortcode_parser.parse(segment['narration_text'])
                segment['narration_text_sanitized_without_shortcodes'] = shortcode_parser.narration_text_sanitized_without_shortcodes
                segment['narration_text_with_simplified_shortcodes'] = shortcode_parser.narration_text_with_simplified_shortcodes
                segment['narration_text_sanitized'] = shortcode_parser.narration_text_sanitized

            if segment['duration'] and isinstance(segment['duration'], str):
                # TODO: Make this consts and/or Enum
                if segment['duration'] == 'SHORTCODE_CONTENT':
                    segment['duration'] = 99997
                elif segment['duration'] == 'FILE_DURATION':
                    segment['duration'] = 99998
                elif segment['duration'] == 'SEGMENT_DURATION':
                    segment['duration'] = 99999
    
    def process_unfinished_projects(self):
        db_projects = self.get_unfinished_projects()

        for db_project in db_projects:
            print_in_progress(f'Processing project "{db_project["_id"]}"')
            self.process_project(db_project)
            print_completed(f'Project "{db_project["_id"]}" processed succesfully!')

    def process_project(self, db_project):
        project = Project(db_project['_id'], db_project['status'], db_project['segments'])
        project.build(f'{self.__projects_output_abspath}project_{project.id}.mp4')
    
    # TODO: This below is for testing, remove it
    def get_unfinished_project(self):
        """
        Returns the first unfinished project that exists in the
        database, or None if there are no unfinished projects.
        """
        return self.__database_handler.get_unfinished_project()

    def get_unfinished_projects(self):
        """
        Returns a list containing all the existing projects in the 
        database that have not been finished yet, or an empty list
        if there are no unfinished projects.
        """
        return self.__database_handler.get_unfinished_projects()
    
    def __read_project_from_file(self, filename: str):
        """
        Reads the provided 'filename', that should contain the information of
        a project, and validates its content and structure. This method will
        return the data read as a dict or will raise an Exception if something
        is not ok.
        """
        if not filename:
            raise Exception('No "filename" provided.')
        
        if not isinstance(filename, str):
            raise Exception('The "filename" parameter provided is a str.')
        
        json_data = read_json_from_file(filename)

        if not 'segments' in json_data:
            raise Exception('The provided "filename" does not contain the expected data structure.')

        # We validate each segment to be able to store the project
        # This will raise an Exception if something is not ok
        for segment in json_data['segments']:
            self.__segment_validator.validate(segment)

        return json_data

    # TODO: Apply ObjectId type to 'project_id'
    def __get_project_from_database_by_id(self, project_id):
        """
        Gets the project with the provided 'project_id' if existing, or
        None if not.
        """
        return self.__database_handler.get_database_project_from_id(project_id)
    
    # TODO: Apply json type to 'json'
    def __get_project_from_database_by_json(self, json):
        """
        Gets the project with the provided 'json' if existing, or None if
        not. This 'json' will be compared with the 'script' field stored
        in database to check if there is a similar project previously 
        stored that must be recovered instead of duplicating it.
        """
        return self.__database_handler.get_database_project_from_json(json)



from youtube_autonomous.shortcodes.enums import ShortcodeType
from youtube_autonomous.shortcodes.objects.shortcode_tag import ShortcodeTag
from youtube_autonomous.shortcodes.objects.shortcode import Shortcode
from yta_general_utils.text.transformer import fix_unseparated_periods, fix_excesive_blank_spaces, fix_separated_parenthesis, add_missing_spaces_before_and_after_parenthesis, fix_separated_square_brackets, add_missing_spaces_before_and_after_square_brackets
from shortcodes import Parser

import re


class ShortcodeParser:
    """
    This class parses the shortcodes from a text, classifies
    them and is capable of handling the information inside.

    We only accept simple shortcodes '[tag]' and block-scoped
    shortcodes '[block_tag] something inside [/block_tag]. 
    The parsing process will detect any invalid shortcode and
    raise an exception.

    _Developer needs to manually register the shortcode parser
    method for each shortcode that this parser will accept._
    """
    __parser: Parser = None
    """
    The shortcodes library parser object.
    """
    text: str = ''
    """
    The last text that this parser object has parsed, as it
    was originally.
    """
    text_sanitized: str = ''
    """
    The last text that this parser object has parsed, but
    sanitized. Any double blank space, any missing blank
    space after a parenthesis, a period or a square bracket
    has been added. Sanitization process is needed to have
    no problems when finding the shortcodes and detecting
    them through the entire text.

    This is the text that is actually used in the parsing
    process to look for the different shortcodes.
    """
    text_sanitized_with_simplified_shortcodes: str = ''
    """
    The last text that this parser object has parsed, with
    the shortcodes simplified after the parsing process
    has detected them. The text used has been sanitized
    previously.
    """
    text_sanitized_without_shortcodes: str = ''
    """
    The last text that this parser object has parsed, with
    no shortcodes on it. The text used has been sanitized
    previously.
    """
    shortcodes: list[Shortcode] = []
    """
    The shortcodes found by the parser in the last text
    parsed.
    """

    def __init__(self, shortcode_tags: list[ShortcodeTag]):
        # TODO: Check that 'shortcode_tags' elements are ShortcodeTags
        self.__parser = Parser(start = '[', end = ']', inherit_globals = False)

        # We register each shortcode in provided 'shortcode_tags'
        self.handlers = []
        for shortcode_tag in shortcode_tags:
            # We have a different handler for each ShortcodeTag type
            if shortcode_tag.is_block_scoped():
                if shortcode_tag.tag == 'blocke':
                    self.__parser.register(self.__blocke_shortcode_handler, shortcode_tag.tag, '/' + shortcode_tag.tag)
            else:
                # TODO: Please, make this dynamic but I couldn't in the past
                if shortcode_tag.tag == 'meme':
                    self.__parser.register(self.__meme_shortcode_handler, shortcode_tag.tag)

        # Wildcards seems to be not working here so we cannot handle
        # any unknown shortcode by ourselves. It is better to raise
        # the exception when an unknown shortcode is found because it
        # means that something is wrong with the next (could be user
        # mitstake or our own mistake)

    def parse(self, text: str):
        """
        This method parses the provided 'text' according to the shortcode
        tags passed in the initializing process. This method will return
        the original text, the text without shortcodes and the list of
        shortcodes found with their indexes and attributes.
        """
        if not text:
            raise Exception('No "text" provided.')
        
        self.text = text
        self.text_sanitized = fix_excesive_blank_spaces(fix_unseparated_periods(fix_separated_parenthesis(fix_separated_square_brackets(add_missing_spaces_before_and_after_parenthesis(add_missing_spaces_before_and_after_square_brackets(text))))))
        self.text_sanitized_with_simplified_shortcodes = ''
        self.text_sanitized_without_shortcodes = ''
        self.shortcodes = []

        # At the begining we have full shortcodes containing attributes
        # but when parsing ([tag attribute=value]), we extract that 
        # information to Shortcode objects and we just preserve the 
        # simple tag ([tag]) to know that there is a shortcode in that
        # position.
        #
        # Once we've done that, we can iterate over the text again,
        # obtaining each shortcode position and extracting the 
        # previous word position in the text.
        
        # We parse shortcodes and extract the information
        self.text_sanitized_with_simplified_shortcodes = self.__parser.parse(self.text_sanitized, context = None)
        
        # After this parse we have 'self.shortcodes' fulfilled

        words = self.text_sanitized_with_simplified_shortcodes.split(' ')

        # Regex to detect any shortcode (start and end tags)
        shortcode_regex_pattern = r'\[(/?[a-zA-Z0-9-_]+)\]'

        # We will iterate over all 
        index = 0
        while index < len(words):
            word = words[index]
            if re.fullmatch(shortcode_regex_pattern, word):
                # TODO: Improve this to avoid the ones completed
                for shortcode in self.shortcodes:
                    if '/' in word:
                        # End shortcode
                        if shortcode.tag == word.replace('[', '').replace(']', '').replace('/', '') and not shortcode.previous_end_word_index:
                            # Is that shortcode
                            shortcode.previous_end_word_index = index - 1
                            break
                    else:
                        # Start shortcode
                        if shortcode.tag == word.replace('[', '').replace(']', '') and not shortcode.previous_start_word_index:
                            # Is that shortcode
                            shortcode.previous_start_word_index = index - 1
                            break

                # TODO: Remove that word
                del(words[index])
            else:
                index += 1

        self.text_sanitized_without_shortcodes = ' '.join(words)

        return self.text_sanitized_without_shortcodes
    
    """
    Specific shortcode handlers below
    TODO: Please, make them work dynamically so I don't need to create 
    a new method in code to any new shortcode, that is stupid.
    """

    # TODO: We need to manually register all shortcode handlers because
    # of some shared code that doesn't allow to do this dynamically (at
    # least by now, there is a task to improve this)
    # TODO: Remove this one, it is a test
    def __meme_shortcode_handler(self, pargs, kwargs, context):
        return self.__simple_shortcode_base_handler(pargs, kwargs, context, 'meme')
    
    def __simple_shortcode_base_handler(self, pargs, kwargs, context, tag: str):
        """
        Handles a simple [tag] shortcode.
        """
        attributes = []

        for parg in pargs:
            attributes.append({
                parg: None
            })

        for kwarg in kwargs:
            attribute_name = kwarg
            attribute_value = kwargs[attribute_name]

            attributes.append({
                attribute_name: attribute_value
            })

        # TODO: We need to handle the position within the text
        self.shortcodes.append(Shortcode(tag, ShortcodeType.SIMPLE, attributes, context, None))
            
        # This will replace the shortcode in the text
        return '[' + tag + ']'

    # TODO: Remove this one, it is a test
    def __blocke_shortcode_handler(self, pargs, kwargs, context, content):
        return self.__block_shortcode_base_handler(pargs, kwargs, context, content, 'blocke')

    def __block_shortcode_base_handler(self, pargs, kwargs, context, content, tag: str):
        """
        Handles a block [tag] ... [/tag] shortcode.
        """
        attributes = []

        for parg in pargs:
            attributes.append({
                parg: None
            })

        for kwarg in kwargs:
            attribute_name = kwarg
            attribute_value = kwargs[attribute_name]

            attributes.append({
                attribute_name: attribute_value
            })

        # TODO: We need to handle the position within the text
        self.shortcodes.append(Shortcode(tag, ShortcodeType.BLOCK, attributes, context, content))
            
        # This will replace the shortcode and its content in the text
        return '[' + tag + ']' + content + '[/' + tag + ']'


from yta_general_utils.programming.parameter_validator import ParameterValidator
from youtube_autonomous.elements.rules.element_rules import ElementRules
from youtube_autonomous.elements.rules.ai_image_element_rules import AIImageElementRules
from youtube_autonomous.elements.rules.image_element_rules import ImageElementRules
from youtube_autonomous.elements.rules.meme_element_rules import MemeElementRules
from youtube_autonomous.elements.rules.ai_video_element_rules import AIVideoElementRules
from youtube_autonomous.elements.rules.video_element_rules import VideoElementRules
from youtube_autonomous.elements.rules.sound_element_rules import SoundElementRules
from youtube_autonomous.elements.rules.custom_stock_element_rules import CustomStockElementRules
from youtube_autonomous.elements.rules.stock_element_rules import StockElementRules
from youtube_autonomous.elements.rules.youtube_video_element_rules import YoutubeVideoElementRules
from youtube_autonomous.elements.rules.text_element_rules import TextElementRules
from youtube_autonomous.elements.rules.premade_element_rules import PremadeElementRules
from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
from youtube_autonomous.elements.validator.error_message import ErrorMessage
from yta_general_utils.file.filename import FileType
from typing import Union


class ElementParameterValidator(ParameterValidator):
    @classmethod
    def validate_keywords(cls, keywords: str):
        return cls.validate_string_mandatory_parameter('keywords', keywords)

    @classmethod
    def validate_text(cls, text: str):
        return cls.validate_string_mandatory_parameter('text', text)
    
    @classmethod
    def validate_premade_name(cls, premade_name: str):
        # TODO: Maybe validate that is a key of a valid premade (?)
        return cls.validate_string_mandatory_parameter('premade_name', premade_name)
    
    @classmethod
    def validate_text_class_name(cls, text_class_name: str):
        return cls.validate_string_mandatory_parameter('text_class_name', text_class_name)
    
    @classmethod
    def validate_duration(cls, duration: Union[int, float]):
        # TODO: Maybe validate that is a key of a valid text class (?)
        return cls.validate_numeric_positive_mandatory_parameter('duration', duration)
    
    @classmethod
    def validate_url(cls, url: str, is_mandatory: True):
        """
        A mandatory 'url' must be not None, a valid string and also
        a valid url and accessible.

        A non mandatory 'url' can be None, 
        """
        if not is_mandatory and not url:
            return True
        
        cls.validate_mandatory_parameter('url', url)
        cls.validate_string_parameter('url', url)
        cls.validate_url_is_ok(url)
        # TODO: Validate 'url' file is some FileType

    @classmethod
    def validate_filename(cls, filename: str, file_type: FileType = FileType.IMAGE, is_mandatory: bool = True):
        """
        Validates the 'filename' parameter. If the 'filename' parameter has
        some value, all conditions will be checked. If there is not 'filename'
        and 'is_mandatory' is False, no Exceptions will be raised because it
        is valid.
        """
        if not is_mandatory and not filename:
            return True
        
        cls.validate_mandatory_parameter('filename', filename)
        cls.validate_string_parameter('filename', filename)
        cls.validate_file_exists('filename', filename)
        cls.validate_filename_is_type('filename', filename, file_type)
        # TODO: Validate type trying to instantiate it

    @classmethod
    def validate_rules(cls, rules: Union[AIImageElementRules, ImageElementRules, MemeElementRules, AIVideoElementRules, VideoElementRules, SoundElementRules, CustomStockElementRules, StockElementRules, YoutubeVideoElementRules, TextElementRules, PremadeElementRules]):
        if not isinstance(rules, ElementRules.get_rules_classes()):
            raise Exception(ErrorMessage.parameter_is_not_rules('rules'))
        
    @classmethod
    def validate_builder(cls, builder: Union[ImageElementBuilder, AIImageElementBuilder, VideoElementBuilder, AIVideoElementBuilder, CustomStockElementBuilder, StockElementBuilder, SoundElementBuilder, MemeElementBuilder, YoutubeVideoElementBuilder, TextElementBuilder, PremadeElementBuilder]):
        if not isinstance(builder, ElementBuilder.get_subclasses()):
            raise Exception(f'The provided "builder" parameter is not a valid ElementBuilder subclass. The valid ones are: {', '.join(ElementBuilder.get_subclasses().__str__)}.')


from youtube_autonomous.segments.enhancement.edition_manual.edition_manual_term import EditionManualTerm
from typing import Union


class EditionManual:
    @property
    def terms(self):
        return self._terms
    
    @terms.setter
    def terms(self, terms: list[EditionManualTerm]):
        if not terms:
            terms = []

        if any(not isinstance(term, EditionManualTerm) for term in terms):
            raise Exception('At least one of the provided "terms" is not an EditionManualTerm.')

        self._terms = terms

    def __init__(self, terms: Union[list[dict], list[EditionManualTerm]]):
        # TODO: We need an ID or something maybe
        self.terms = terms


from youtube_autonomous.segments.enhancement.enhancement_element import EnhancementElement, EnhancementElementMode
from yta_multimedia.video.utils import parse_parameter_as_moviepy_clip
from yta_multimedia.greenscreen.custom.image_greenscreen import ImageGreenscreen
from yta_multimedia.greenscreen.custom.video_greenscreen import VideoGreenscreen
from yta_general_utils.downloader.google_drive import download_file_from_google_drive
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip
from typing import Union


class GreenscreenEnhancementElement(EnhancementElement):
    @classmethod
    def get_valid_modes(cls):
        """
        The EnhancementElementMode enums accepted for this type of
        EnhancementElement.

        See this link: https://www.notion.so/Segment-types-configuration-107f5a32d46280beb793d84a2af7f75e?pvs=4
        """
        return [EnhancementElementMode.REPLACE]

    @classmethod
    def get_default_mode(cls):
        """
        The EnhancementElementMode enum defined as value by default.
        """
        return EnhancementElementMode.REPLACE

    @property
    def valid_modes(self):
        """
        The EnhancementElementMode enums accepted for this type of
        EnhancementElement.

        See this link: https://www.notion.so/Segment-types-configuration-107f5a32d46280beb793d84a2af7f75e?pvs=4
        """
        return GreenscreenEnhancementElement.get_valid_modes()

    def _build(self, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
        filename = ''
        if self.filename:
            filename = self.filename
        elif self.url:
            # TODO: This won't work, I need to modify this method to
            # accept different type of urls and analyze them
            # TODO: By now, accept only Google Drive urls
            filename = download_file_from_google_drive(self.url)
        elif self.keywords:
            # TODO: Look for green screens in a new channel that I must
            # create about greenscreen backgrounds only
            # TODO: By now I'm forcing this ImageScreen
            filename = download_file_from_google_drive('https://drive.google.com/file/d/1WQVnXY1mrw-quVXOqTBJm8x9scEO_JNz/view?usp=drive_link')

        if file_is_video_file(filename):
            greenscreen = VideoGreenscreen(filename)
            # TODO: If 'filename' duration is less than 'video' we
            # should, at least, let the user know that we will make
            # our own modification strategy
        else:
            greenscreen = ImageGreenscreen(filename)

        super().recalculate_duration(None, video)

        end = self.start + self.duration
        enhancement_clip = greenscreen.from_video_to_video(video.subclip(self.start, end))

        return enhancement_clip
    
    def apply(self, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
        """
        This method will apply this Enhancement Element in the provided
        'video' and will return it modified including the generated
        enhancement clip.
        """
        video = parse_parameter_as_moviepy_clip(video)

        enhancement_clip = self._build(video)
        
        video = super().apply(enhancement_clip, video)

        return video





from youtube_autonomous.elements.element_builder import ElementBuilder
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.enums import SegmentField
from yta_multimedia.image.edition.resize import resize_image_scaling
from yta_multimedia.video.generation import generate_video_from_image
from yta_general_utils.downloader.image import download_image
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.temp import create_temp_filename
from typing import Union


class ImageElementBuilder(ElementBuilder):
    """
    This builders allows you to generate 'IMAGE' content for the
    segments.
    """
    @classmethod
    def build_from_filename(cls, filename: str, duration: Union[float, int]):
        ElementParameterValidator.validate_filename(filename)
        ElementParameterValidator.validate_duration(duration)

        return cls.image_to_video(filename, duration, None)

    @classmethod
    def build_from_url(cls, url: str, duration: Union[float, int]):
        ElementParameterValidator.validate_url(url)
        ElementParameterValidator.validate_duration(duration)

        # TODO: Try to do all this process in memory, writting not the image
        image_filename = download_image(url)

        return cls.image_to_video(image_filename, duration, None)
    
    @classmethod
    def build_from_filename_from_segment(cls, segment: dict):
        filename = segment.get(SegmentField.FILENAME.value, None)
        duration = segment.get('calculated_duration', None)

        return cls.build_from_filename(filename, duration)

    @classmethod
    def build_from_url_from_segment(cls, segment: dict):
        url = segment.get(SegmentField.URL.value, None)
        duration = segment.get('calculated_duration', None)

        return cls.build_from_url(url, duration)
    
    # TODO: Move this to a helper (?)
    @classmethod
    def image_to_video(cls, filename: str, duration: Union[float, int], effect = None):
        ElementParameterValidator.validate_filename(filename)
        ElementParameterValidator.validate_duration(duration)
        
        # Resize image to fit the screen
        tmp_image_filename = create_temp_filename(f'image{get_file_extension(filename)}')
        resize_image_scaling(filename, 1920, 1080, tmp_image_filename)

        # TODO: Apply and Effect (this need work)
        video = generate_video_from_image(tmp_image_filename, duration, effect)

        return video